#include <math.h>
#ifndef PI
#define PI (atan(1)*4)
#endif
#ifndef StoTENbits
#define StoTENbits 1023.0/724.0
#endif
#ifndef HtoTENbits
#define HtoTENbits 1023.0/360.0
#endif
#ifndef R_StoTENbits
#define R_StoTENbits 724.0/1023.0
#endif
#ifndef R_HtoTENbits
#define R_HtoTENbits 360.0/1023.0
#endif

//	CMYK\
typedef	struct {
	unsigned char	c,m,y,k;		//	f(CMYK)f[^ (0`255)
} PIXEL_CMYK;

//	HSV\
typedef	struct {
	short	h;		//	f( F )f[^ ( 0 ` 1023  )	͈͊OɏoĂ邱Ƃz肵ĂȂBF10bitɂtɈς?
	short	s;		//	f( ʓx )f[^ ( 0 ` 1023  )	͈͊OɏoĂ邱Ƃz肵ĂȂ
	short	v;		//	f( x )f[^ ( 0 ` 4095 )	YC\̂ŷ܂ܑBĔ͈͊OɏoĂ邱Ƃ
} PIXEL_HSV;

void colorspace_convert_table();
void rgb2cmyk( PIXEL_CMYK* cmykp, PIXEL* pixelp, int h, int w, int skip );
void cmyk2rgb( PIXEL* pixelp, PIXEL_CMYK* cmykp, int h, int w, int skip );
void yc2hsv( PIXEL_YC* ycp, PIXEL_HSV* hsvp, int h, int w, int skip );
void hsv2yc( PIXEL_HSV* hsvp, PIXEL_YC* ycp, int h, int w, int skip );

static unsigned char	CMYK2RGB[256][256];
static unsigned char	RGB2CMYK[256][256];
static short			YC2H[1024][1024];
static short			YC2S[513][513];
static short			HS2Cr[1024][1024];
static short			HS2Cb[1024][1024];

void colorspace_convert_table()
{
	int i,j,temp;
	//RGBCMYKϊ̎OvZ
	for(i=0;i<256;i++){
		RGB2CMYK[i][255]=0;
		for(j=0;j<255;j++){
			temp = int( (255-i-j)/(double)((255-j))*255 );
			RGB2CMYK[i][j] = (temp < 0) ? 0 : ((temp > 255) ? 255 : temp);
		}
	}
	//CMYKRGBϊ̎OvZ
	for(i=0;i<256;i++){
		for(j=0;j<255;j++){
			temp = int( (i/255.0*(255-j)-255+j)*-1 );
			CMYK2RGB[i][j] = (temp < 0) ? 0 : ((temp > 255) ? 255 : temp);
		}
	}
	//CbCrHOvZ
	for(i=0;i<1024;i++){
		for(j=0;j<1024;j++){
			if(i==512 && j==512){
				YC2H[i][j] = 0;	//CbCr0Ȃ΁AƂ肠H0Ă
			} else {
				temp = int((atan2((double)(j-512),(double)(i-512))*180.0/PI) *HtoTENbits);
				YC2H[i][j] = short( (temp < 0) ? temp+1023 : ((temp > 1023) ? temp-1023 : temp) );
			}
		}
	}
	//CbCrSOvZ
	for(i=0;i<513;i++){
		for(j=0;j<513;j++){
			temp = int(sqrt((double)(i*i+j*j))*StoTENbits);
			YC2S[i][j]=short((temp < 0) ? 0 : ((temp > 1023) ? 1023 : temp));
		}
	}
	//HSCbCrOvZ
	for(i=0;i<1024;i++){
		for(j=0;j<1024;j++){
			HS2Cb[i][j] = short(j*R_StoTENbits*cos(PI*(double)(i*R_HtoTENbits)/180.0) *4);
			HS2Cr[i][j] = short(j*R_StoTENbits*sin(PI*(double)(i*R_HtoTENbits)/180.0) *4);
		}
	}
}

void rgb2cmyk( PIXEL_CMYK* cmykp, PIXEL* pixp, int h, int w, int skip )
{
	int x,y;
	for(y=0;y<h;y++){
		for(x=0;x<w;x++){
			if(pixp->r >= pixp->g && pixp->r >= pixp->b)
				cmykp->k = 255-pixp->r;
			else if(pixp->g >= pixp->r && pixp->g >= pixp->b)
				cmykp->k = 255-pixp->g;
			else
				cmykp->k = 255-pixp->b;
			cmykp->c = RGB2CMYK[pixp->r][cmykp->k];
			cmykp->m = RGB2CMYK[pixp->g][cmykp->k];
			cmykp->y = RGB2CMYK[pixp->b][cmykp->k];
			cmykp++;
			pixp++;
		}
		cmykp += skip;
		pixp += skip;
	}
}

void cmyk2rgb( PIXEL* pixp, PIXEL_CMYK* cmykp, int h, int w, int skip )
{
	int x,y;
	for(y=0;y<h;y++){
		for(x=0;x<w;x++){
			pixp->r = CMYK2RGB[cmykp->c][cmykp->k];
			pixp->g = CMYK2RGB[cmykp->m][cmykp->k];
			pixp->b = CMYK2RGB[cmykp->y][cmykp->k];
			cmykp++;
			pixp++;
		}
		cmykp += skip;
		pixp += skip;
	}
}

void yc2hsv( PIXEL_HSV* hsvp, PIXEL_YC* ycp, int h, int w, int skip )
{
	int x,y;
	for(y=0;y<h;y++){
		for(x=0;x<w;x++){
			//-----z͈͊OANZXh~
			if((ycp->cb + 2048) & 0xf000)
				ycp->cb = (ycp->cb < -2048) ? -2048 : 2047;
			if((ycp->cr + 2048) & 0xf000)
				ycp->cr = (ycp->cr < -2048) ? -2048 : 2047;
			ycp->cb >>=2;
			ycp->cr >>=2;

			hsvp->h = YC2H[ ycp->cb+512 ][ ycp->cr+512 ];
			hsvp->s = YC2S[ (ycp->cb<0) ? 0-ycp->cb : ycp->cb ][ (ycp->cr<0) ? 0-ycp->cr : ycp->cr ];
			hsvp->v = ycp->y;
			hsvp++;
			ycp++;
		}
		hsvp += skip;
		ycp += skip;
	}
}

void hsv2yc( PIXEL_YC* ycp, PIXEL_HSV* hsvp, int h, int w, int skip )
{
	int x,y;
	for(y=0;y<h;y++){
		for(x=0;x<w;x++){
/*
		//-----z͈͊OANZXh~
		if(hsvp->h < 0)	hsvp->h = 0;
			else if(hsvp->h > 1023)	hsvp->h = 1023;
		if(hsvp->s < 0)	hsvp->s = 0;
			else if(hsvp->s > 1023)	hsvp->s = 1023;
*/
		ycp->cb = HS2Cb[hsvp->h][hsvp->s];
		ycp->cr = HS2Cr[hsvp->h][hsvp->s];
		ycp->y = hsvp->v;
		hsvp++;
		ycp++;
		}
		hsvp += skip;
		ycp += skip;
	}
}
